import os

from flask import Flask
from flask_cors import CORS

##__________________________________________________________________||
def create_app(config_path=None):

    app = Flask(__name__, instance_relative_config=False)
    app.config.from_mapping(
        SECRET_KEY='dev',
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        SQLALCHEMY_TRACK_MODIFICATIONS=False
    )

    if config_path is not None:
        if not os.path.isabs(config_path):
            # If `config_path` is relative, it is considered relative
            # to the current working directory.
            config_path = os.path.join(os.getcwd(), config_path)

            # Note: `app.config.from_pyfile()` treats a relative path
            # either a) relative to the top directory of the app,
            # i.e., the directory in which this file is, if
            # `instance_relative_config` is `False`, or b) relative to
            # the "instance folder", a specific directory described in
            # https://flask.palletsprojects.com/en/1.1.x/config/#instance-folders),
            # if `instance_relative_config` is `True`.

        app.config.from_pyfile(config_path, silent=False)

    from . import db
    db.init_app(app)

    from . import bpquery
    bpquery.init_app(app)

    from . import bpgraphql
    bpgraphql.init_app(app)

    CORS(app, resources={r'/*': {'origins': '*'}})

    return app

##__________________________________________________________________||

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType, SQLAlchemyConnectionField

from .models import Map as MapModel
from .models import Beam as BeamModel
from .models import MapFilePath as MapFilePathModel

##__________________________________________________________________||
class Map(SQLAlchemyObjectType):
    class Meta:
        model = MapModel
        interfaces = (relay.Node, )

class MapConnection(relay.Connection):
    class Meta:
        node = Map

##__________________________________________________________________||
class Beam(SQLAlchemyObjectType):
    class Meta:
        model = BeamModel
        interfaces = (relay.Node, )

# class BeamConnection(relay.Connection):
#      class Meta:
#          node = Beam

## The class "BeamConnection" is commented out because it causes the error
## "AssertionError: Found different types with the same name in the schema:
## BeamConnection, BeamConnection". In the class "Query" below
## "Beam._meta.connection" is used instead. This solution was mentioned in the
## issue
## https://github.com/graphql-python/graphene-sqlalchemy/issues/153#issuecomment-478744077

##__________________________________________________________________||
class MapFilePath(SQLAlchemyObjectType):
    class Meta:
        model = MapFilePathModel
        interfaces = (relay.Node, )

# class MapFilePathConnection(relay.Connection):
#     class Meta:
#         node = MapFilePath

##__________________________________________________________________||
class Query(graphene.ObjectType):
    node = relay.Node.Field()
    all_maps = SQLAlchemyConnectionField(MapConnection)
    all_beams = SQLAlchemyConnectionField(Beam._meta.connection)
    all_map_file_paths = SQLAlchemyConnectionField(MapFilePath)

schema = graphene.Schema(query=Query, types=[Map, Beam, MapFilePath])

##__________________________________________________________________||



import os
import datetime
import json
import csv

import click
from flask import current_app, g
from flask.cli import with_appcontext
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import MetaData
import sqlalchemy

##__________________________________________________________________||
db = SQLAlchemy()

##__________________________________________________________________||
def get_db_connection():
    if 'db_connection' not in g:
        g.db_connection = db.engine.connect()
    return g.db_connection

##__________________________________________________________________||
def close_db_connection(e=None):
    conn = g.pop('db_connection', None)
    if conn is not None:
        conn.close()

##__________________________________________________________________||
def init_app(app):
    db.init_app(app)
    app.cli.add_command(init_db_command)
    app.cli.add_command(dump_db_command)
    app.cli.add_command(import_csv_command)
    app.teardown_appcontext(close_db_connection)

##__________________________________________________________________||
def init_db():

    engine = db.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)

    if metadata.tables:
        tbl_names = metadata.tables.keys()
        # ['maps', 'beams']

        tbl_names = ', '.join(['"{}"'.format(t) for t in tbl_names])
        # '"beams", "maps"'

        msg = "Dropped all tables: {}".format(tbl_names)

        metadata.drop_all(bind=engine)
        print(msg)

    if not db.Model.metadata.tables:
        msg = "No tables to be created are found!"
        print(msg)
        return

    tbl_names = db.Model.metadata.tables.keys()
    # ['maps', 'beams']

    tbl_names = ', '.join(['"{}"'.format(t) for t in tbl_names])
    # '"beams", "maps"'

    msg = "Created tables: {}".format(tbl_names)

    db.Model.metadata.create_all(engine)

    print(msg)

@click.command("init-db")
@with_appcontext
def init_db_command():
    """Clear existing data and create new tables."""
    init_db()
    click.echo("Initialized the database.")

##__________________________________________________________________||
def get_all_db_content():
    # https://stackoverflow.com/questions/47307873/read-entire-database-with-sqlalchemy-and-dump-as-json
    engine = db.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)
    ret = { }
    for tbl in metadata.sorted_tables:
        ret[tbl.name] = [dict(r) for r in engine.execute(tbl.select())]
    return ret

@click.command("dump-db")
@with_appcontext
def dump_db_command():
    db_content = get_all_db_content()

    click.echo(json.dumps(db_content, indent=2, default=str))
    # https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable

##__________________________________________________________________||
def import_csv(csvdir):
    metadata = MetaData()
    metadata.reflect(bind=db.engine)
    for tbl in metadata.sorted_tables:
        csv_filename = '{}.csv'.format(tbl.name)
        csv_path = os.path.join(csvdir, csv_filename)
        if os.path.exists(csv_path):
            import_csv_(tbl, csv_path)
            message = 'imported to "{}" from {}'.format(tbl.name, csv_path)
        else:
            message = 'skipped "{}". file not found: {}'.format(tbl.name, csv_path)
        print(message)

def import_csv_(tbl, csv_path):
    with open(csv_path, 'r') as f:
        rows = list(csv.reader(f))
    fields = rows[0]
    rows = rows[1:]
    data = [{f: convert_type(e, tbl.columns[f].type) for f, e in zip(fields, r)} for r in rows]
    ins = tbl.insert()
    connection = get_db_connection()
    connection.execute(ins, data)

def convert_type(str_, type_):
    if isinstance(type_, sqlalchemy.sql.sqltypes.DATE):
        if str_:
            return datetime.datetime.strptime(str_, "%Y-%m-%d").date()
        return None
    return str_

@click.command("import-csv")
@click.argument("csvdir")
@with_appcontext
def import_csv_command(csvdir):
    import_csv(csvdir)

##__________________________________________________________________||



from .db import db

##__________________________________________________________________||
class Map(db.Model):
    __tablename__ = 'maps'
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.Text(), nullable=False, unique=True, index=True)
    date_posted = db.Column(db.Date())
    mapper = db.Column(db.Text())
    note = db.Column(db.Text())

class Beam(db.Model):
    __tablename__ = 'beams'
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.Text(), nullable=False, unique=True, index=True)
    path = db.Column(db.Text())
    input_map_id = db.Column(db.ForeignKey('maps.id'))
    input_beam_id = db.Column(db.ForeignKey('beams.id'))
    map = db.relationship("Map", backref=db.backref("beams"))
    parent_beam = db.relationship(lambda: Beam, remote_side=id, backref=db.backref("child_beams"))

class MapFilePath(db.Model):
    __tablename__ = 'map_path'
    id = db.Column(db.Integer(), primary_key=True)
    map_id = db.Column(db.ForeignKey('maps.id'))
    path = db.Column(db.Text())
    note = db.Column(db.Text())
    map = db.relationship("Map", backref=db.backref("map_file_paths"))

##__________________________________________________________________||
